unit OutputStreams;

interface

uses Windows, SysUtils, Types, Classes, IniFiles;

type
  TLineOutputStream = class(TStream)
  protected
    Ini: TCustomIniFile;
    LineLength: integer;
    LineBuffer: PByte;
    LineBufferPos: PByte;
    LineBufferUsed: integer;
    LineBufferUnUsed: integer;
    procedure AddLine(); virtual;
  public
    constructor Create(LineLength: integer);
    destructor Destroy; override;
    function Write(const Buffer; Count: longint): longint; override;
    procedure Finalize();
  end;

  TIniFileOutputStream = class(TLineOutputStream)
  protected
    FTotalBytesWritten: integer;
    FLineCount: integer;
    IniFile:    TCustomIniFile;
    IniSection: string;
    procedure AddLine(); override;
  public
    constructor Create(IniFile: TCustomIniFile; IniSection: string;
      LineLength: integer);
    function Write(const Buffer; Count: longint): longint; override;
    property TotalBytesWritten: integer Read FTotalBytesWritten;
    property LineCount: integer Read FLineCount;
  end;

  TStringsOutputStream = class(TLineOutputStream)
    FTotalBytesWritten: integer;
    FLineCount: integer;
    Strings:    TStrings;
    procedure AddLine(); override;
  public
    constructor Create(Strings: TStrings; LineLength: integer);
    function Write(const Buffer; Count: longint): longint; override;
    property TotalBytesWritten: integer Read FTotalBytesWritten;
    property LineCount: integer Read FLineCount;
  end;

  TWin32FileStream = class(THandleStream)
  private
    FFileName: string;
  public
    //See Win32 API function CreateFile for description of parameters
    constructor Create(FileName: string; DesiredAccess: DWORD;
      ShareMode: DWORD; CreationDisposition: DWORD; FlagsAndAttributes: DWORD);
    destructor Destroy; override;
    property FileName: string Read FFileName;
  end;

implementation

constructor TLineOutputStream.Create(LineLength: integer);
var
  tmp: PByte;
begin
  self.LineLength := LineLength;
  GetMem(Linebuffer, LineLength + 1);

  // Set the last byte to zer -> null terminate for PChar string
  tmp := Linebuffer;
  Inc(tmp, LineLength);
  tmp^ := 0;

  LineBufferUsed   := 0;
  LineBufferUnUsed := LineLength;
  LineBufferPos    := LineBuffer;
end;

destructor TLineOutputStream.Destroy;
begin
  Finalize();
  FreeMem(LineBuffer);
  inherited;
end;

 // Writes the characters from Buffer line by line into the
 // internal LineBuffer. After one line is full it is
 // added to the destination
function TLineOutputStream.Write(const Buffer; Count: longint): longint;
var
  Written: integer;
  Input:   PByte;
begin
  Input   := Pbyte(@Buffer);
  Written := 0;
  while (Count >= LineBufferUnUsed) do
  begin
    System.Move(Input^, LineBufferPos^, LineBufferUnUsed);
    Inc(Written, LineBufferUnUsed);
    Dec(Count, LineBufferUnUsed);
    Inc(Input, LineBufferUnUsed);
    AddLine();
  end;
  if (Count > 0) then
  begin
    System.Move(Input^, LineBufferPos^, Count);
    Inc(Written, Count);
    Inc(LineBufferPos, Count);
    Inc(LineBufferUsed, Count);
    Dec(LineBufferUnUsed, Count);
  end;
  Result := Written;
end;

// Write the last bytes from the internal LineBuffer to destination
procedure TLineOutputStream.Finalize();
begin
  if (LineBufferUsed = 0) then
    exit;
  LineBufferPos^ := 0;
  addLine();
end;

// Write the current LineBuffer including it's line-number to the destination
procedure TLineOutputStream.AddLine();
begin
  LineBufferPos    := LineBuffer;
  LineBufferUsed   := 0;
  LineBufferUnUsed := LineLength;
end;

constructor TIniFileOutputStream.Create(IniFile: TCustomIniFile;
  IniSection: string; LineLength: integer);
begin
  inherited Create(LineLength);
  self.IniFile    := IniFile;
  self.IniSection := IniSection;
  FTotalBytesWritten := 0;
  FLineCount      := 0;
end;

function TIniFileOutputStream.Write(const Buffer; Count: longint): longint;
begin
  Result := inherited Write(Buffer, Count);
  Inc(FTotalBytesWritten, Result);
end;

procedure TIniFileOutputStream.AddLine();
begin
  IniFile.WriteString(IniSection, IntToStr(FLineCount), PChar(LineBuffer));
  Inc(FLineCount);
  inherited;
end;

constructor TStringsOutputStream.Create(Strings: TStrings; LineLength: integer);
begin
  inherited Create(LineLength);
  self.Strings := Strings;
  FTotalBytesWritten := 0;
  FLineCount   := 0;
end;

function TStringsOutputStream.Write(const Buffer; Count: longint): longint;
begin
  Result := inherited Write(Buffer, Count);
  Inc(FTotalBytesWritten, Result);
end;

procedure TStringsOutputStream.AddLine();
begin
  Strings.Add(Format('%d=%s', [FLineCount, PChar(LineBuffer)]));
  Inc(FLineCount);
  inherited;
end;

constructor TWin32FileStream.Create(FileName: string; DesiredAccess: DWORD;
  ShareMode: DWORD; CreationDisposition: DWORD; FlagsAndAttributes: DWORD);
begin
  inherited Create(CreateFile(PChar(FileName), DesiredAccess, ShareMode,
    nil, CreationDisposition, FlagsAndAttributes, 0));
  if (FHandle = 0) then
    RaiseLastWin32Error;
  self.FFileName := FileName;
end;

destructor TWin32FileStream.Destroy;
begin
  if FHandle >= 0 then
    FileClose(FHandle);
  inherited Destroy;
end;


end.
