unit ZLIBArchive2;

//***********************************************
// ZLBArchive Control.  Ver. 2.0
//***********************************************

interface

uses
  Windows, SysUtils, Classes, ZLibEx, ZLIBArchiveGlobals2;

type
  TZLBArc2 = class
  private
    { Private declarations }
    fNewArchive: boolean;
    fArchive: string;
    fArchiveDetails: TZLBArchiveRec2;
    fArchiveRegistry: TList;
    fCompressionLevel: TFileCompressionLevel;
    fExtractWithPath: boolean;
    fRecursePaths: boolean;
    fSavePaths: boolean;
    fVerifyFileCRC: boolean;
    fProgressFlag: TInternalFlags;
    fProgressText: string;
    fProgressTotal : int64;
    fProgressIndex : int64;
    fVerifyArchiveCRC: boolean;
    fMemoryUsage: integer;
    FTempFolder: string;
    fPassword : string;
    fZeroPos : Int64;
    fVersionReadonly : boolean;

    fOnReplaceFile: TReplaceFileEvent;
    fOnFileExtracted: TFileExtractedEvent;
    fOnStreamExtracted: TFileExtractedEvent;
    fOnArchiveOpened: TNotifyEvent;
    fOnErrorLog: TErrorLogEvent;
    fOnProgress: TProgressEvent;
    fOnFileAddUpdate: TFileAddUpdateEvent;
    fOnDecryptStream: TStreamDecryptEvent;
    fOnEncryptStream: TStreamEncryptEvent;
    function GetFileCount: longint;
    function GetFileItem(index: longint): TZLBFileRec2;
    function Extract(Index: longint; OutFile: TStream; InStrm: TStream = nil): boolean;
    function Add(InFileName: string; OutFile: TStream; encrypt:boolean = false): TZLBFileRec2;
    function Update(InFileName: string; OutFile: TStream; filerec:TZLBFileRec2; encrypt:boolean = false): TZLBFileRec2;
    procedure InternalProgress(Sender: TObject);
    procedure UpdateRegistryRec(FileIndex: longint; FileInfo: TZLBFileRec2);
    procedure WriteRegistry(DataStrm: TStream);
    function GetTempFolder: string;
  protected
    { Protected declarations }
    function GetTmpFileName: string; virtual;
    procedure SearchTree(path: string; var files: TStrings); virtual;
    function InternalFilePos(ReportedPos:Int64):Int64;
  public
    { Public declarations }
    constructor create();
    destructor Destroy; override;

    property ArchiveName: string read fArchive;
    function OpenArchive(filename: string): boolean;
    procedure CreateArchive(filename: string);
    procedure CloseArchive;

    procedure ConvertArchive(NewFilename:string);

    procedure AddFile(Filename: string; encrypt:boolean = false);
    procedure AddFiles(FileList: TStrings; encrypt:boolean = false);
    procedure AddStream(Filename: string; InStream: TStream; encrypt:boolean = false);
    procedure DeleteFiles(FileList: TStrings);
    procedure ExtractAll(Destination: string);
    procedure ExtractFileByIndex(Destination: string; index: longint);
    procedure ExtractStreamByIndex(OutFile: TStream; index: longint);
    procedure ExtractFileByName(Destination, Filename: string);
    procedure ExtractStreamByName(OutFile: TStream; Filename: string);
    procedure UpdateFile(FileName: string; FileIndex: longint; encrypt:boolean = false);
    procedure UpdateFiles(FileList: TStrings; encrypt:boolean = false);
    procedure UpdateStream(FileName: string; InStream: TStream; encrypt:boolean = false);
    procedure GetFileList(FileList: TStrings);

    procedure RenameFileByName(Original, New: string);
    function FileInArchive(filename: string): longint;

    property FileCount: longint read GetFileCount;
    property Files[index: longint]: TZLBFileRec2 read GetFileItem;

    property TempWorkingFolder: string read GetTempFolder write fTempFolder;
    property CryptPassword:string read fPassword write fPassword;
    property VersionReadonly : boolean read fVersionReadonly;
  published
    { Published declarations }
    property MaxMemoryUsage: integer read fMemoryUsage write fMemoryUsage default 20;
    property CompressionLevel: TFileCompressionLevel read FCompressionLevel write FCompressionLevel default fcDefault;
    property ExtractWithPath: boolean read fExtractWithPath write fExtractWithPath default false;
    property VerifyArchiveCRC: boolean read fVerifyArchiveCRC write fVerifyArchiveCRC default true;
    property VerifyFileCRC: boolean read fVerifyFileCRC write fVerifyFileCRC default false;
    property SavePaths: boolean read fSavePaths write fSavePaths default false;
    property RecursePaths: boolean read fRecursePaths write fRecursePaths default false;
    property OnArchiveOpened: TNotifyEvent read fOnArchiveOpened write fOnArchiveOpened;
    property OnFileAddUpdate: TFileAddUpdateEvent read fOnFileAddUpdate write fOnfileaddUpdate;
    property OnReplaceFile: TReplaceFileEvent read fOnReplaceFile write fOnReplaceFile;
    property OnFileExtracted: TFileExtractedEvent read fOnFileExtracted write fOnFileExtracted;
    property OnEncryptStream: TStreamEncryptEvent read fOnEncryptStream write fOnEncryptStream;
    property OnDecryptStream: TStreamDecryptEvent read fOnDecryptStream write fOnDecryptStream;
    property OnStreamExtracted: TFileExtractedEvent read fOnStreamExtracted write fOnStreamExtracted;
    property OnErrorLogged: TErrorLogEvent read fOnErrorLog write fOnErrorLog;
    property OnProgress: TProgressEvent read fOnProgress write fOnProgress;
  end;


implementation

uses consts;

constructor TZLBArc2.create();
begin
  inherited create();
  fNewArchive := false;
  FCompressionLevel := fcdefault;
  fExtractWithPath := false;
  fArchive := '';
  fArchiveRegistry := TList.create;
  fVerifyFileCRC := false;
  fSavePaths := false;
  fRecursePaths := false;
  fMemoryUsage := 20;
  fVerifyArchiveCRC := true;
  fTempFolder := '';
  fZeroPos := -1;
end;

destructor TZLBArc2.destroy;
begin
  if farchive <> '' then
    CloseArchive;
  fArchiveRegistry.free;
  inherited;
end;

function TZLBArc2.OpenArchive(filename: string): boolean;
var
  InFile: TFileStream;
  TmpFile: TStream;
  ZStream: TCustomZStream;
  loop, count: longint;
  Data: ^TZLBFileRec2;
  OldData : TZLBFileRec;
  CheckValue, fsize, buffersize: longint;
  buffer: pointer;
  wArchiveDetails : TZLBArchiveRec;
begin
  CloseArchive;
  fArchive := FileName;
  result := false;
  if (FileName = '') then
  begin
    fArchive := '';
    exit;
  end;

  if not FileExists(FileName) then
    raise EFileNotFound.create(ArchiveFileNotFound);

  try
    InFile := TFileStream.Create(FileName, fmOpenRead + fmShareDenyWrite);
  except
    raise EGeneralArchiveException.create(format(UnableToOpenFile, [filename]));
  end;

  try
    InFile.Seek(-sizeof(TZLBArchiveRec2), soFromEnd);
    InFile.read(fArchiveDetails, sizeof(TZLBArchiveRec2));

    if fArchiveDetails.Version = 1 then
    begin
       fVersionReadonly := true;
       InFile.Seek(-sizeof(TZLBArchiveRec), soFromEnd);
       InFile.read(wArchiveDetails, sizeof(TZLBArchiveRec));
       fillchar(fArchiveDetails, sizeof(TZLBArchiveRec2), 0);
       
       fArchiveDetails.CRC := wArchiveDetails.CRC;
       fArchiveDetails.FileCount := wArchiveDetails.FileCount;
       fArchiveDetails.Version := wArchiveDetails.Version;
       fArchiveDetails.RegistrySize := wArchiveDetails.RegistrySize;
       fArchiveDetails.ArchiveSize := wArchiveDetails.ArchiveSize;
       fArchiveDetails.CompressedRegistry := wArchiveDetails.CompressedRegistry;
       fArchiveDetails.ArchiveEncrypted := false;

    end
    else if fArchiveDetails.Version = 2 then
       fVersionReadonly := false
    else
    raise EVersionError.create(format(ZLBArchiveVersionError, [fArchiveDetails.Version]));

    Infile.Seek(-(fArchiveDetails.ArchiveSize + fArchiveDetails.RegistrySize + sizeof(TZLBArchiveRec2)), soFromEnd);
    fZeroPos := InFile.Position;

    if fVerifyArchiveCRC then
    begin
      fProgressText := VerifyingArchive;
      CheckValue := GetCRC32(InFile, sizeof(TZLBArchiveRec2), InternalFilePos(0));
    end
    else
      checkvalue := 0;

    if fVerifyArchiveCRC and (fArchiveDetails.CRC <> CheckValue) then
    begin
      fArchive := '';
      fillchar(fArchiveDetails, sizeof(TZLBArchiveRec2), 0);
      raise EFileCorruption.create(ArchiveCorrupted);
    end;

    try
      if not fArchiveDetails.CompressedRegistry then
      begin
        InFile.Seek(InternalFilePos(fArchiveDetails.ArchiveSize), soFromBeginning);
        for loop := 0 to fArchiveDetails.FileCount - 1 do
        begin
          Data := nil;

          if fArchiveDetails.Version = 2 then
          begin
             getmem(Data, sizeof(TZLBFileRec2));
             InFile.read(Data^, sizeof(TZLBFileRec2));
          end
          else if fArchiveDetails.Version = 1 then
          begin
             InFile.read(OldData, sizeof(TZLBFileRec));

             getmem(Data, sizeof(TZLBFileRec2));
             Data^.name := OldData.Name;
             Data^.path := OldData.path;
             Data^.osize := OldData.osize;
             Data^.csize := OldData.csize;
             Data^.esize := 0;
             Data^.start := OldData.start;
             data^.CRC := OldData.CRC;
             Data^.Status := OldData.Status;
             Data^.CompressionLevel := OldData.CompressionLevel;
          end;

          if data <> nil then
             fArchiveRegistry.add(Data);
        end;
      end
      else
      begin
        InFile.Seek(InternalFilePos(fArchiveDetails.ArchiveSize), soFromBeginning);
        tmpfile := tmemorystream.Create;
        try
          if fArchiveDetails.Version = 2 then
             tmemorystream(tmpfile).SetSize(sizeof(TZLBFileRec2) * fArchiveDetails.filecount)
          else if fArchiveDetails.Version = 1 then
             tmemorystream(tmpfile).SetSize(sizeof(TZLBFileRec) * fArchiveDetails.filecount);

          ZStream := TZDecompressionStream.Create(InFile);
          try
            fsize := 0;
            
            if fArchiveDetails.Version = 2 then
               fsize := sizeof(TZLBFileRec2) * fArchiveDetails.filecount
            else if fArchiveDetails.Version = 1 then
               fsize := sizeof(TZLBFileRec) * fArchiveDetails.filecount;

            fProgressFlag := ifOpening;
            fProgressTotal := fSize;
            fProgressIndex := 0;
            fProgressText := ExtractingRegistry;
            while True do
            begin
              if fsize <= 0 then
                break;
              if fsize >= 4096 then
                buffersize := 4096
              else
                buffersize := fsize;
              getmem(buffer, buffersize);
              try
                try
                  Count := ZStream.Read(Buffer^, BufferSize);
                  inc(fProgressIndex, Count);
                  InternalProgress(self);
                  if Count <> 0 then
                    TmpFile.WriteBuffer(Buffer^, Count);
                except
                  raise EArchiveRegistry.create(RegistryReadError);
                end;
              finally
                freemem(buffer, buffersize);
              end;
              dec(fsize, buffersize);
            end;
          finally
            zstream.free;
          end;
          tmpfile.seek(0, soFromBeginning);
          for loop := 0 to fArchiveDetails.filecount - 1 do
          begin
             Data := nil;

             if fArchiveDetails.Version = 2 then
             begin
                getmem(Data, sizeof(TZLBFileRec2));
                TmpFile.read(Data^, sizeof(TZLBFileRec2));
             end
             else if fArchiveDetails.Version = 1 then
             begin
                TmpFile.read(OldData, sizeof(TZLBFileRec));

                getmem(Data, sizeof(TZLBFileRec2));
                Data^.name := OldData.Name;
                Data^.path := OldData.path;
                Data^.osize := OldData.osize;
                Data^.csize := OldData.csize;
                Data^.esize := 0;
                Data^.start := OldData.start;
                data^.CRC := OldData.CRC;
                Data^.Status := OldData.Status;
                Data^.CompressionLevel := OldData.CompressionLevel;
             end;

             if Data <> nil then
                fArchiveRegistry.add(Data);
          end;
        finally
          tmpfile.free;
        end
      end;
      if assigned(fOnArchiveOpened) then
        fOnArchiveOpened(self);
    except
      CloseArchive;
      raise EArchiveRegistry.create(RegistryReadError);
    end;
  finally
    InFile.free;
  end;
  result := true;
  fProgressFlag := ifClear;
  InternalProgress(self);
end;

procedure TZLBArc2.CloseArchive;
var
  data: pointer;
  loop: longint;
begin
  for loop := fArchiveRegistry.Count - 1 downto 0 do
  begin
    data := fArchiveRegistry.Items[loop];
    fArchiveRegistry.Delete(loop);
    freemem(data, sizeof(TZLBFileRec2));
  end;
  fNewArchive := false;
  fArchiveRegistry.Clear;
  fillchar(fArchiveDetails, sizeof(TZLBArchiveRec2), 0);
  fArchive := '';
  fProgressFlag := ifClear;
  InternalProgress(self);
  fZeroPos := -1;
end;

function TZLBArc2.GetFileCount: longint;
begin
  if fArchive <> '' then
    Result := fArchiveRegistry.count
  else
    raise ENoArchive.create(NoArchiveOpen);
end;

function TZLBArc2.GetFileItem(index: longint): TZLBFileRec2;
begin
  if (index >= 0) and (index < fArchiveRegistry.count) then
    result := TZLBFileRec2(fArchiveRegistry.items[index]^)
  else
    raise EArchiveIndexBounds.create(IndexBounds);
end;

procedure TZLBArc2.ExtractAll(Destination: string);
var
  InFile: TStream;
  OutFile: TStream;
  data: TZLBFileRec2;
  loop: longint;
  successfull: boolean;
  extractionpath: string;
  overwrite: boolean;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  try
    InFile := TFileStream.Create(farchive, fmOpenRead + fmShareDenyWrite)
  except
    raise EGeneralArchiveException.create(format(UnableToOpenFile, [fArchive]));
  end;

  try
    for loop := 0 to filecount - 1 do
    begin
      successfull := false;
      data := Files[loop];

      extractionpath := destination;
      if ExtractWithPath then
        extractionpath := extractionpath + data.path;

      while pos('\\', extractionpath) > 0 do
        delete(extractionpath, pos('\\', extractionpath), 1);

      if extractionpath[length(extractionpath)] = '\' then
        delete(extractionpath, length(extractionpath), 1);
      ForceDirectories(extractionpath);

      if not DirectoryExists(extractionpath) then
        raise EExtractionError.create(format(FilePathError, [extractionpath]));

      if (fileexists(extractionpath + '\' + data.name) and assigned(fonreplacefile)) then
      begin
        overwrite := false;
        fonreplacefile(extractionpath + '\' + data.name, overwrite);
      end
      else
        overwrite := true;

      if overwrite then
      begin
        OutFile := TFileStream.Create(extractionpath + '\' + data.name, fmCreate);
        try
          try
             successfull := Extract(loop, OutFile, InFile);
          except
             on e:Exception do
             begin
                if assigned(fOnErrorLog) then
                   fOnErrorLog(e.Message);
             end;
          end
        finally
          OutFile.Free;
          if successfull and assigned(fOnFileExtracted) then
            fOnFileExtracted(destination + data.name);
        end;
      end;
    end;
  finally
    InFile.Free;
    fProgressFlag := ifClear;
    InternalProgress(self);
  end;
end;

function TZLBArc2.Extract(Index: longint; OutFile: TStream; InStrm: TStream = nil): boolean;
var
  InFile: TStream;
  CryptStrm : TStream;
  data: TZLBFileRec2;
  successfull: boolean;
  ZStream: TCustomZStream;
  fsize, count, buffersize: longint;
  buffer: pointer;
  tmpfile:string;

procedure InternalDecompress(InStrm:TStream);
begin
   if (data.Status in [fsCompressed, fsCompressedEncrypted]) then
   begin
      ZStream := TZDecompressionStream.Create(InStrm);
      try
        fsize := data.osize;
        fProgressFlag := ifExtracting;
        fProgressTotal := fSize;
        fProgressIndex := 0;
        fProgressText := format(ExtractingFile, [data.name]);
        while True do
        begin
          if fsize <= 0 then
            break;
          if fsize >= 4096 then
            buffersize := 4096
          else
            buffersize := fsize;
          getmem(buffer, buffersize);
          try
            try
              Count := ZStream.Read(Buffer^, BufferSize);
              inc(fProgressIndex, count);
              InternalProgress(self);
              if Count <> 0 then
                OutFile.WriteBuffer(Buffer^, Count);
            except
              raise EFileCorruption.create(ArchiveCorrupted);
            end;
          finally
            freemem(buffer, buffersize);
          end;
          dec(fsize, buffersize);
        end;
      finally
        ZStream.Free;
      end;
   end
   else
   begin
      if data.osize > 0 then
         OutFile.CopyFrom(InStrm, data.osize);
   end;
end;

begin
  successfull := false;
  data := Files[index];

  if not assigned(InStrm) then
  begin
     try
       InFile := TFileStream.Create(farchive, fmOpenRead + fmShareDenyWrite)
     except
       raise EGeneralArchiveException.create(format(UnableToOpenFile, [fArchive]));
     end;
  end
  else
     InFile := InStrm;

  try
    InFile.Seek(InternalFilePos(data.start), soFromBeginning);

    if (data.Status in [fsCompressedEncrypted, fsStoredEncrypted]) then
    begin
       if (trim(fPassword) <> '') then
       begin
          if Assigned(fOnDecryptStream) then
          begin
             tmpfile := GetTmpFileName;
             try
                CryptStrm := TFileStream.Create(tmpfile, fmopenreadwrite + fmshareexclusive);
                try
                   fOnDecryptStream(infile, cryptstrm, data.esize, fPassword);
                   cryptstrm.Position := 0;
                   InternalDecompress(CryptStrm);
                finally
                   CryptStrm.free;
                end;
             finally
                deletefile(tmpfile);
             end;
          end
          else
          raise EExtractionError.create(NoDecryptionDetected);
       end
       else
       raise EExtractionError.create(NoPasswordGiven);
    end
    else
       InternalDecompress(InFile);

    if (VerifyFileCRC) and (GetCRC32(OutFile, 0, 0) <> data.CRC) then
      raise EFileCorruption.create(Format(FileCRC, [data.name]));

    successfull := true;
  finally
    if InStrm = nil then
       InFile.free;
    result := successfull;
  end;
end;

procedure TZLBArc2.ExtractFileByIndex(Destination: string; index: longint);
var
  OutFile: TStream;
  data: TZLBFileRec2;
  extractionpath: string;
  overwrite: boolean;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  if not ((index >= 0) and (index < fArchiveRegistry.count)) then
    raise EArchiveIndexBounds.create(IndexBounds);

  data := Files[index];

  extractionpath := destination;
  if ExtractWithPath then
  begin
    extractionpath := extractionpath + data.path;
    while pos('\\', extractionpath) > 0 do
      delete(extractionpath, pos('\\', extractionpath), 1);
    if extractionpath[length(extractionpath)] = '\' then
      delete(extractionpath, length(extractionpath), 1);
    ForceDirectories(extractionpath);
    if not DirectoryExists(extractionpath) then
      raise EExtractionError.create(format(FilePathError, [extractionpath]));
  end;
  if (fileexists(extractionpath + '\' + data.name) and assigned(fonreplacefile)) then
  begin
    overwrite := false;
    fonreplacefile(extractionpath + '\' + data.name, overwrite);
  end
  else
    overwrite := true;
  if overwrite then
  begin
    OutFile := TFileStream.Create(extractionpath + '\' + data.name, fmCreate);
    try
      if Extract(index, Outfile) and assigned(fOnFileExtracted) then
        fOnFileExtracted(destination + data.name);
    finally
      OutFile.free;
    end;
  end;
  fProgressFlag := ifClear;
  InternalProgress(self);
end;

procedure TZLBArc2.ExtractFileByName(Destination, Filename: string);
var
  OutFile: TStream;
  data: TZLBFileRec2;
  index, loop: longint;
  extractionpath: string;
  overwrite: boolean;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  index := -1;
  for loop := 0 to filecount - 1 do
  begin
    data := Files[loop];
    if lowercase(data.path + data.name) = lowercase(filename) then
    begin
      index := loop;
      break;
    end;
  end;

  if index = -1 then
    raise EExtractionError.create(Format(FileNotFound, [filename]));

  extractionpath := destination;
  if ExtractWithPath then
    extractionpath := extractionpath + data.path;
  while pos('\\', extractionpath) > 0 do
    delete(extractionpath, pos('\\', extractionpath), 1);
  if extractionpath[length(extractionpath)] = '\' then
    delete(extractionpath, length(extractionpath), 1);
  ForceDirectories(extractionpath);
  if not DirectoryExists(extractionpath) then
    raise EGeneralArchiveException.create(format(FilePathError, [extractionpath]));

  if (fileexists(extractionpath + '\' + data.name) and assigned(fonreplacefile)) then
  begin
    overwrite := false;
    fonreplacefile(extractionpath + '\' + data.name, overwrite);
  end
  else
    overwrite := true;
  if overwrite then
  begin
    OutFile := TFileStream.Create(extractionpath + '\' + data.name, fmCreate);
    try
      if Extract(index, Outfile) and assigned(fOnFileExtracted) then
        fOnFileExtracted(destination + data.name);
    finally
      OutFile.free;
    end;
  end;
  fProgressFlag := ifClear;
  InternalProgress(self);
end;

procedure TZLBArc2.ExtractStreamByIndex(OutFile: TStream; index: longint);
var
  data: TZLBFileRec2;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  if (index < 0) and (index >= fArchiveRegistry.count) then
    raise EArchiveIndexBounds.create(IndexBounds);

  data := Files[index];
  if Extract(index, Outfile) and assigned(fOnStreamExtracted) then
    fOnStreamExtracted(data.path + data.name);
  fProgressFlag := ifClear;
  InternalProgress(self);
end;

procedure TZLBArc2.ExtractStreamByName(OutFile: TStream; Filename: string);
var
  data: TZLBFileRec2;
  index, loop: longint;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  index := -1;
  for loop := 0 to filecount - 1 do
  begin
    data := Files[loop];
    if lowercase(data.path + data.name) = lowercase(filename) then
    begin
      index := loop;
      break;
    end;
  end;

  if index = -1 then
    raise EFileNotFoundInArchive.create(FileNotFound);

  data := Files[index];
  if Extract(index, Outfile) and assigned(fOnStreamExtracted) then
    fOnStreamExtracted(data.path + data.name);
  fProgressFlag := ifClear;
  InternalProgress(self);
end;

procedure TZLBArc2.SearchTree(path: string; var files: TStrings);

  procedure ScanPath(Path: string);
  var
    data: TSearchRec;
    ds: longint;
    FoundFiles: tstringlist;
  begin
    if (Path[length(Path)] = '\') then
      delete(Path, length(Path), 1);
    FoundFiles := tstringlist.create;
    ds := findfirst(Path + '\*.*', faAnyFile, data);
    while ds = 0 do
    begin
      if (data.attr and faDirectory = faDirectory) then
      begin
        if (data.name <> '.') and
          (data.name <> '..') then
        begin
          if fRecursePaths then
            ScanPath(Path + '\' + data.name);
          if fSavePaths then
            FoundFiles.add(Path + '\' + data.name);
        end;
      end
      else
        FoundFiles.add(Path + '\' + data.name);
      ds := findnext(data);
    end;
    findClose(data);
    files.addstrings(FoundFiles);
    FoundFiles.Free;
  end;

begin
  scanPath(path);
end;

function TZLBArc2.Add(InFileName:string; OutFile:TStream; Encrypt:boolean = false):TZLBFileRec2;
var
   ZStream: TCustomZStream;
   InFile, TmpStrm: TStream;
   FileInfo: TZLBFileRec2;
   MemStrm: Boolean;
   TmpFile: string;
begin
    try
      InFile := TFileStream.Create(InFileName, fmOpenRead + fmShareDenyWrite);
    except
      raise EGeneralArchiveException.create(format(UnableToOpenFile, [Infilename]));
    end;

    try
      if InFile.Size <= (fMemoryUsage * (1024 * 1024)) then
      begin
        TmpStrm := TMemoryStream.create;
        MemStrm := true;
        TmpFile := '';
      end
      else
      begin
        TmpFile := GetTmpFileName;
        TmpStrm := TFileStream.create(TmpFile, fmCreate);
        MemStrm := false;
      end;

      try
         ZStream := TZCompressionStream.Create(tmpStrm, TZCompressionLevel(fCompressionLevel));
         try
           fProgressFlag := ifAdding;
           fProgressTotal := Infile.Size;
           fProgressIndex := 0;
           fProgressText := format(UpdatingFile, [InFileName]);
           TZCompressionStream(ZStream).OnProgress := InternalProgress;
           ZStream.CopyFrom(InFile, 0);
         finally
           ZStream.Free;
         end;
         TmpStrm.position := 0;

         fillchar(FileInfo, sizeof(TZLBFileRec2), 0);
         Infile.Position := 0;
         FileInfo.CRC := GetCRC32(InFile, 0, 0);
         FileInfo.name := extractfilename(InFileName);
         FileInfo.path := extractfilepath(InFileName);
         if SavePaths then
           delete(FileInfo.path, 1, pos(':', FileInfo.path))
         else
           fillchar(FileInfo.path, sizeof(FileInfo.path), #0);
         FileInfo.osize := InFile.size;
         FileInfo.start := OutFile.Position - fZeroPos;

         if Encrypt and assigned(fOnEncryptStream) then
         begin
            if TmpStrm.Size >= InFile.size then
            begin
              if not MemStrm then
              begin
                 TFileStream(tmpStrm).Free;
                 TmpStrm := TFileStream.Create(TmpFile, fmCreate);
              end
              else
                 TMemoryStream(tmpStrm).Clear;
                 
              tmpStrm.Position := 0;
              InFile.Position := 0;
              fOnEncryptStream(InFile, tmpstrm, fPassword);
              OutFile.CopyFrom(TmpStrm, 0);
              FileInfo.esize := tmpstrm.size;
              FileInfo.Status := fsStoredEncrypted;
              FileInfo.CompressionLevel := fcNone;
            end
            else
            begin
              fOnEncryptStream(tmpstrm, outfile, fpassword);
              FileInfo.CSize := TmpStrm.Size;
              FileInfo.esize := OutFile.Position - InternalFilePos(FileInfo.start);
              FileInfo.Status := fsCompressedEncrypted;
              FileInfo.CompressionLevel := fCompressionLevel;
            end;
         end
         else
         begin
            if TmpStrm.Size >= InFile.size then
            begin
              OutFile.CopyFrom(InFile, 0);
              FileInfo.Status := fsStored;
              FileInfo.CompressionLevel := fcNone;
            end
            else
            begin
              OutFile.CopyFrom(TmpStrm, 0);
              FileInfo.CSize := TmpStrm.Size;
              FileInfo.Status := fsCompressed;
              FileInfo.CompressionLevel := fCompressionLevel;
            end;
         end;
      finally
        TmpStrm.free;
        if not MemStrm then
        begin
          DeleteFile(TmpFile);
          tmpfile := '';
        end;
      end;

    finally
      InFile.free;
    end;

    result := FileInfo;
end;

procedure TZLBArc2.AddFile(Filename: string; encrypt:boolean = false);
var
  ArchiveFile, NewFile : TStream;
  TmpFileName: string;
  FileInfo: ^TZLBFileRec2;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  if fVersionReadonly then
    raise EVersionError.create(OlderArchiveFormat);

  TmpFileName := '';
  if savepaths then
    TmpFileName := extractfilepath(filename);
  TmpFileName := TmpFileName + ExtractFilename(filename);

  if fileinarchive(TmpFileName) > -1 then
    raise EGeneralArchiveException.create(Format(FileAlreadyExists, [TmpFileName]));

  TmpFileName := GetTmpFileName;
  NewFile := TFileStream.Create(TmpFileName, fmOpenReadWrite + fmShareExclusive);
  try
    if not fnewarchive then
    begin
      try
        ArchiveFile := TFileStream.Create(fArchive, fmOpenRead + fmShareDenyWrite);
      except
        raise EGeneralArchiveException.create(format(UnableToOpenFile, [fArchive]));
      end;
      try
        ArchiveFile.Position := 0;
        NewFile.CopyFrom(ArchiveFile, InternalFilePos(fArchiveDetails.ArchiveSize));
      finally
        ArchiveFile.Free;
      end;
    end;

    GetMem(FileInfo, sizeof(TZLBFileRec2));
    FileInfo^ := Add(filename, NewFile, encrypt);
    fArchiveRegistry.add(FileInfo);

    if assigned(fOnFileAddUpdate) then
      fOnFileAddUpdate(FileInfo^.path, FileInfo^.Name);

    WriteRegistry(NewFile);

    NewFile.Free;
    DeleteFile(fArchive);

    fProgressFlag := ifMovingTempFile;
    fProgressText := MovingTempFile;
    InternalProgress(self);

    RenameFile(TmpFileName, fArchive);
    TmpFileName := fArchive;
    CloseArchive;
    OpenArchive(TmpFileName);
  except
    DeleteFile(TmpFileName)
  end;
end;

procedure TZLBArc2.AddFiles(FileList: TStrings; encrypt:boolean = false);
var
  ArchiveFile, NewFile: TStream;
  TmpFileName, TestName: string;
  FileInfo: ^TZLBFileRec2;
  Loop: longint;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  if fVersionReadonly then
    raise EVersionError.create(OlderArchiveFormat);

  loop := 0;
  while loop < Filelist.Count do
  begin
    if directoryexists(FileList[loop]) then
    begin
      if RecursePaths then
        searchtree(FileList[loop], filelist);
      filelist.Delete(loop);
    end
    else
      inc(loop);
  end;

  TmpFileName := GetTmpFileName;
  NewFile := TFileStream.Create(TmpFileName, fmOpenReadWrite + fmShareExclusive);
  try
    try
      if not fnewarchive then
      begin
        ArchiveFile := TFileStream.Create(fArchive, fmOpenRead + fmShareDenyWrite);
        try
          ArchiveFile.Position := 0;
          NewFile.CopyFrom(ArchiveFile, InternalFilePos(fArchiveDetails.ArchiveSize));
        finally
          ArchiveFile.Free;
        end;
      end;
    except
      raise EGeneralArchiveException.create(format(UnableToOpenFile, [fArchive]));
    end;

    while FileList.count > 0 do
    begin
      TestName := ExtractFilePath(filelist[0]);
      if SavePaths then
        delete(TestName, 1, pos(':', TestName))
      else
        TestName := '';
      TestName := TestName + ExtractFilename(FileList[0]);

      if fileinarchive(TestName) > -1 then
        if assigned(fOnErrorLog) then
          fOnErrorLog(format(FileAlreadyExists, [FileList[0]]));

      try
         GetMem(FileInfo, sizeof(TZLBFileRec2));
         FileInfo^ := Add(filelist[0], NewFile, encrypt);
         fArchiveRegistry.add(FileInfo);

         if assigned(fOnFileAddUpdate) then
           fOnFileAddUpdate(FileInfo^.path, FileInfo^.Name);
      except
        if assigned(fOnErrorLog) then
          fOnErrorLog(format(UnableToOpenFile, [FileList[0]]));
      end;

      FileList.Delete(0);
    end;
    WriteRegistry(NewFile);
    NewFile.Free;
    DeleteFile(fArchive);
    fProgressFlag := ifMovingTempFile;
    fProgressText := MovingTempFile;
    InternalProgress(self);
    RenameFile(TmpFileName, fArchive);
    TmpFileName := fArchive;
    CloseArchive;
    OpenArchive(TmpFileName);
  except
    DeleteFile(TmpFileName)
  end;
end;

procedure TZLBArc2.AddStream(Filename: string; InStream: TStream; encrypt:boolean = false);
var
  ArchiveFile, NewFile, TmpStrm: TStream;
  ZStream: TCustomZStream;
  TmpFileName: string;
  FileInfo: ^TZLBFileRec2;
  MemStrm: Boolean;
  TmpFile: string;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  if fVersionReadonly then
    raise EVersionError.create(OlderArchiveFormat);

  TmpFileName := '';
  if savepaths then
    TmpFileName := extractfilepath(filename);
  TmpFileName := TmpFileName + ExtractFilename(filename);

  if fileinarchive(TmpFileName) > -1 then
    raise EGeneralArchiveException.create(FileAlreadyExists);

  TmpFileName := GetTmpFileName;
  NewFile := TFileStream.Create(TmpFileName, fmOpenReadWrite + fmShareExclusive);
  try
    try
      if not fnewarchive then
      begin
        ArchiveFile := TFileStream.Create(fArchive, fmOpenRead + fmShareExclusive);
        try
          ArchiveFile.Position := 0;
          NewFile.CopyFrom(ArchiveFile, InternalFilePos(fArchiveDetails.ArchiveSize));
        finally
          ArchiveFile.Free;
        end;
      end;
    except
      raise EGeneralArchiveException.create(format(UnableToOpenFile, [fArchive]));
    end;

    if InStream.Size <= (fMemoryUsage * (1024 * 1024)) then
    begin
      TmpStrm := TMemoryStream.create;
      MemStrm := true;
      TmpFile := '';
    end
    else
    begin
      TmpFile := GetTmpFileName;
      TmpStrm := TFileStream.create(TmpFile, fmCreate);
      MemStrm := false;
    end;
    try
      ZStream := TZCompressionStream.Create(TmpStrm, TZCompressionLevel(fCompressionLevel));
      try
        fProgressFlag := ifAdding;
        fProgressTotal := InStream.Size;
        fProgressIndex := 0;
        fProgressText := format(UpdatingFile, [filename]);
        TZCompressionStream(ZStream).OnProgress := InternalProgress;
        ZStream.CopyFrom(InStream, 0);
      finally
        ZStream.Free;
      end;
      TmpStrm.position := 0;

      getmem(FileInfo, sizeof(TZLBFileRec2));
      fillchar(FileInfo^, sizeof(TZLBFileRec2), 0);
      FileInfo^.CRC := GetCRC32(InStream, 0, 0);
      FileInfo^.name := extractfilename(filename);
      FileInfo^.path := extractfilepath(filename);
      if SavePaths then
        delete(FileInfo^.path, 1, pos(':', FileInfo^.path))
      else
        fillchar(FileInfo^.path, sizeof(FileInfo^.path), #0);

      FileInfo^.osize := InStream.size;
      FileInfo^.start := NewFile.Position - fZeroPos;

      if Encrypt and assigned(fOnEncryptStream) then
      begin
         if TmpStrm.Size >= InStream.size then
         begin

           if not MemStrm then
           begin
              TFileStream(tmpStrm).Free;
              TmpStrm := TFileStream.Create(TmpFile, fmCreate);
           end
           else
              TMemoryStream(tmpStrm).Clear;

           tmpStrm.Position := 0;
           InStream.Position := 0;
           fOnEncryptStream(InStream, tmpstrm, fPassword);
           NewFile.CopyFrom(TmpStrm, 0);
           FileInfo^.esize := tmpstrm.Size;
           FileInfo^.Status := fsStoredEncrypted;
           FileInfo^.CompressionLevel := fcNone;
         end
         else
         begin
           fOnEncryptStream(tmpstrm, Newfile, fpassword);
           FileInfo^.CSize := TmpStrm.Size;
           FileInfo^.esize := NewFile.Position - InternalFilePos(FileInfo.start);
           FileInfo^.Status := fsCompressedEncrypted;
           FileInfo^.CompressionLevel := fCompressionLevel;
         end;
      end
      else
      begin
         if TmpStrm.Size >= InStream.size then
         begin
           NewFile.CopyFrom(InStream, 0);
           FileInfo^.Status := fsStored;
           FileInfo^.CompressionLevel := fcNone;
         end
         else
         begin
           NewFile.CopyFrom(TmpStrm, 0);
           FileInfo^.CSize := TmpStrm.Size;
           FileInfo^.Status := fsCompressed;
           FileInfo^.CompressionLevel := fCompressionLevel;
         end;
      end;

    finally
      TmpStrm.free;
      if not MemStrm then
      begin
        DeleteFile(TmpFile);
        tmpfile := '';
      end;
    end;

    fArchiveRegistry.add(FileInfo);

    if assigned(fOnFileAddUpdate) then
      fOnFileAddUpdate(FileInfo^.path, FileInfo^.Name);

    WriteRegistry(NewFile);

    NewFile.Free;
    DeleteFile(fArchive);

    fProgressFlag := ifMovingTempFile;
    fProgressText := MovingTempFile;
    InternalProgress(self);
    RenameFile(TmpFileName, fArchive);

    TmpFileName := fArchive;
    CloseArchive;
    OpenArchive(TmpFileName);
  except
    DeleteFile(TmpFileName)
  end;
end;

function TZLBArc2.Update(InFileName: string; OutFile: TStream; filerec:TZLBFileRec2; encrypt: boolean): TZLBFileRec2;
var
   ZStream: TCustomZStream;
   InFile, TmpStrm: TStream;
   FileInfo: TZLBFileRec2;
   MemStrm: Boolean;
   TmpFile: string;
begin
   try
     InFile := TFileStream.Create(InFileName, fmOpenRead + fmShareDenyWrite);
   except
     raise EGeneralArchiveException.create(format(UnableToOpenFile, [InFileName]));
   end;

   try
     if InFile.Size <= (fMemoryUsage * (1024 * 1024)) then
     begin
       TmpStrm := TMemoryStream.create;
       MemStrm := true;
       TmpFile := '';
     end
     else
     begin
       TmpFile := GetTmpFileName;
       TmpStrm := TFileStream.create(TmpFile, fmCreate);
       MemStrm := false;
     end;
     try
       ZStream := TZCompressionStream.Create(TmpStrm, TZCompressionLevel(fCompressionLevel));
       try
         fProgressFlag := ifUpdating;
         fProgressTotal := Infile.Size;
         fProgressIndex := 0;
         fProgressText := format(UpdatingFile, [InFileName]);
         TZCompressionStream(ZStream).OnProgress := InternalProgress;
         ZStream.CopyFrom(InFile, 0);
       finally
         ZStream.Free;
       end;
       TmpStrm.position := 0;
       InFile.Position := 0;

       fillchar(FileInfo, sizeof(TZLBFileRec2), 0);
       FileInfo.name := filerec.name;
       FileInfo.path := filerec.path;
       FileInfo.CRC := GetCRC32(InFile, 0, 0);
       FileInfo.osize := Infile.Size;
       FileInfo.start := OutFile.Position - fZeroPos;

       if Encrypt and assigned(fOnEncryptStream) then
       begin
          if TmpStrm.Size >= InFile.size then
          begin
            if not MemStrm then
            begin
               TFileStream(tmpStrm).Free;
               TmpStrm := TFileStream.Create(TmpFile, fmCreate);
            end
            else
               TMemoryStream(tmpStrm).Clear;

            tmpStrm.Position := 0;
            InFile.Position := 0;
            fOnEncryptStream(InFile, tmpstrm, fPassword);
            OutFile.CopyFrom(TmpStrm, 0);
            FileInfo.esize := tmpstrm.Size;
            FileInfo.Status := fsStoredEncrypted;
            FileInfo.CompressionLevel := fcNone;
          end
          else
          begin
            fOnEncryptStream(tmpstrm, outfile, fpassword);
            FileInfo.CSize := TmpStrm.Size;
            FileInfo.esize := OutFile.Position - InternalFilePos(FileInfo.start);
            FileInfo.Status := fsCompressedEncrypted;
            FileInfo.CompressionLevel := fCompressionLevel;
          end;
       end
       else
       begin
          if TmpStrm.Size >= InFile.size then
          begin
            OutFile.CopyFrom(InFile, 0);
            FileInfo.Status := fsStored;
            FileInfo.CompressionLevel := fcNone;
          end
          else
          begin
            OutFile.CopyFrom(TmpStrm, 0);
            FileInfo.CSize := TmpStrm.Size;
            FileInfo.Status := fsCompressed;
            FileInfo.CompressionLevel := fCompressionLevel;
          end;
       end;
     finally
       TmpStrm.free;
       if not MemStrm then
       begin
         DeleteFile(TmpFile);
         tmpfile := '';
       end;
     end;
   finally
     InFile.free;
   end;
   result := fileinfo;
end;

procedure TZLBArc2.UpdateFile(FileName: string; FileIndex: longint; encrypt:boolean = false);
var
  ArchiveFile, NewFile: TStream;
  TmpFileName: string;
  FileInfo: TZLBFileRec2;
  loop: longint;
  wSize : Int64;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  if fVersionReadonly then
    raise EVersionError.create(OlderArchiveFormat);

  TmpFileName := '';
  if savepaths then
    TmpFileName := extractfilepath(filename);
  TmpFileName := TmpFileName + ExtractFilename(filename);

  if fileinarchive(TmpFileName) <> FileIndex then
    raise EGeneralArchiveException.create(FileNotFound);

  FileInfo := Files[FileIndex];

  TmpFileName := GetTmpFileName;
  NewFile := TFileStream.Create(TmpFileName, fmOpenReadWrite + fmShareExclusive);
  try
    try
      ArchiveFile := TFileStream.Create(fArchive, fmOpenRead + fmShareDenyWrite);
    except
      raise EGeneralArchiveException.create(format(UnableToOpenFile, [fArchive]));
    end;

    try
      ArchiveFile.Position := 0;
      if FileInfo.Start > 0 then
        NewFile.CopyFrom(ArchiveFile, InternalFilePos(FileInfo.start));

      if FileInfo.eSize > 0 then
         wSize := FileInfo.eSize
      else if fileInfo.csize > 0 then
         wSize := FileInfo.cSize
      else
         wSize := fileinfo.osize;

      ArchiveFile.Position := ArchiveFile.Position + wSize;

      fileinfo := Update(filename, newfile, fileinfo, encrypt);

      UpdateRegistryRec(FileIndex, FileInfo);

      if assigned(fOnFileAddUpdate) then
        fOnFileAddUpdate(FileInfo.path, FileInfo.Name);

      for loop := FileIndex + 1 to FileCount - 1 do
      begin
        FileInfo := files[loop];
        FileInfo.start := NewFile.Position-fZeroPos;
        UpdateRegistryRec(loop, FileInfo);

        if FileInfo.eSize > 0 then
           wSize := FileInfo.eSize
        else if fileInfo.csize > 0 then
           wSize := FileInfo.cSize
        else
           wSize := fileinfo.osize;

        if wSize > 0 then
          NewFile.CopyFrom(ArchiveFile, wSize);
      end;

      WriteRegistry(NewFile);

    finally
      ArchiveFile.free;
    end;
    NewFile.Free;
    DeleteFile(fArchive);

    fProgressFlag := ifMovingTempFile;
    fProgressText := MovingTempFile;
    InternalProgress(self);

    RenameFile(TmpFileName, fArchive);
    TmpFileName := fArchive;
    CloseArchive;
    OpenArchive(TmpFileName);
  except
    DeleteFile(TmpFileName)
  end;
end;

procedure TZLBArc2.UpdateFiles(FileList: TStrings; encrypt:boolean = false);
var
  ArchiveFile, NewFile: TStream;
  TmpFileName, TestName: string;
  FileInfo: TZLBFileRec2;
  loop, loop2: longint;
  FileFound: boolean;
  wSize : Int64;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  if fVersionReadonly then
    raise EVersionError.create(OlderArchiveFormat);

  loop := 0;
  while loop < Filelist.Count do
  begin
    if directoryexists(FileList[loop]) then
    begin
      if RecursePaths then
        searchtree(FileList[loop], filelist);
      filelist.Delete(loop);
    end
    else
      inc(loop);
  end;

  TmpFileName := GetTmpFileName;
  NewFile := TFileStream.Create(TmpFileName, fmOpenReadWrite + fmShareExclusive);
  try
    try
      ArchiveFile := TFileStream.Create(fArchive, fmOpenRead + fmShareDenyWrite);
    except
      raise EGeneralArchiveException.create(format(UnableToOpenFile, [fArchive]));
    end;
    try
      loop := 0;

      if fZeroPos > 0 then
         NewFile.CopyFrom(ArchiveFile, fZeroPos);

      while loop < filecount do
      begin
        FileInfo := Files[loop];
        loop2 := 0;
        FileFound := false;
        while loop2 < filelist.count do
        begin
          TestName := extractfilepath(filelist[loop2]);
          if SavePaths then
            delete(TestName, 1, pos(':', TestName))
          else
            TestName := '';
          TestName := TestName + extractfilename(filelist[loop2]);
          if lowercase(files[loop].path + files[loop].name) = lowercase(TestName) then
          begin
            filefound := true;
            break;
          end;
          inc(loop2);
        end;

        if filefound then
        begin //updating file from original archive with new file.
          try

            fileinfo := Update(filelist[loop2], newfile, fileinfo, encrypt);

            UpdateRegistryRec(loop, FileInfo);

            if assigned(fOnFileAddUpdate) then
              fOnFileAddUpdate(FileInfo.path, FileInfo.Name);

          except
            raise EGeneralArchiveException.create(format(UnableToOpenFile, [fArchive]));
          end;
        end
        else
        begin //Moving file from original archive.
          ArchiveFile.Seek(InternalFilePos(FileInfo.start), soFromBeginning);

          FileInfo.start := NewFile.Position-fZeroPos;

          if FileInfo.eSize > 0 then
             wSize := FileInfo.eSize
          else if fileInfo.csize > 0 then
             wSize := FileInfo.cSize
          else
             wSize := fileinfo.osize;

          if (wSize > 0) then
            NewFile.CopyFrom(ArchiveFile, wSize);
            
          UpdateRegistryRec(loop, FileInfo);
        end;
        inc(loop);
      end;

      WriteRegistry(NewFile);

    finally
      ArchiveFile.free;
    end;
    NewFile.Free;
    DeleteFile(fArchive);

    fProgressFlag := ifMovingTempFile;
    fProgressText := MovingTempFile;
    InternalProgress(self);

    RenameFile(TmpFileName, fArchive);
    TmpFileName := fArchive;
    CloseArchive;
    OpenArchive(TmpFileName);
  except
    DeleteFile(TmpFileName)
  end;
end;

procedure TZLBArc2.UpdateStream(FileName: string; InStream: TStream; encrypt:boolean = false);
var
  ArchiveFile, NewFile, TmpStrm: TStream;
  ZStream: TCustomZStream;
  TmpFileName: string;
  FileInfo: TZLBFileRec2;
  loop, FileIndex: longint;
  MemStrm: Boolean;
  TmpFile: string;
  wSize : Int64;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  if fVersionReadonly then
    raise EVersionError.create(OlderArchiveFormat);

  TmpFileName := '';
  if savepaths then
    TmpFileName := extractfilepath(filename);
  TmpFileName := TmpFileName + ExtractFilename(filename);

  FileIndex := fileinarchive(TmpFileName);
  if FileIndex = -1 then
    raise EGeneralArchiveException.create(FileNotFound);

  FileInfo := Files[FileIndex];

  TmpFileName := GetTmpFileName;
  NewFile := TFileStream.Create(TmpFileName, fmOpenReadWrite + fmShareExclusive);
  try
    try
      ArchiveFile := TFileStream.Create(fArchive, fmOpenRead + fmShareDenyWrite);
    except
      raise EGeneralArchiveException.create(format(UnableToOpenFile, [fArchive]));
    end;

    try
      ArchiveFile.Position := 0;

      if FileInfo.Start > 0 then
        NewFile.CopyFrom(ArchiveFile, InternalFilePos(FileInfo.start));

      if FileInfo.eSize > 0 then
         wSize := FileInfo.eSize
      else if fileInfo.csize > 0 then
         wSize := FileInfo.cSize
      else
         wSize := fileinfo.osize;

      ArchiveFile.Position := ArchiveFile.Position + wSize;

      if InStream.Size <= (fMemoryUsage * (1024 * 1024)) then
      begin
        TmpStrm := TMemoryStream.create;
        MemStrm := true;
        TmpFile := '';
      end
      else
      begin
        TmpFile := GetTmpFileName;
        TmpStrm := TFileStream.create(TmpFile, fmCreate);
        MemStrm := false;
      end;
      try
        ZStream := TZCompressionStream.Create(TmpStrm, TZCompressionLevel(fCompressionLevel));
        try
          fProgressFlag := ifUpdating;
          fProgressTotal := InStream.Size;
          fProgressIndex := 0;
          fProgressText := format(UpdatingFile, [filename]);
          TZCompressionStream(ZStream).OnProgress := InternalProgress;
          ZStream.CopyFrom(InStream, 0);
        finally
          ZStream.Free;
        end;
        TmpStrm.position := 0;
        FileInfo.CRC := GetCRC32(InStream, 0, 0);
        FileInfo.osize := InStream.Size;

        if Encrypt and assigned(fOnEncryptStream) then
        begin
           if TmpStrm.Size >= InStream.size then
           begin
             if not MemStrm then
             begin
                TFileStream(tmpStrm).Free;
                TmpStrm := TFileStream.Create(TmpFile, fmCreate);
             end
             else
                TMemoryStream(tmpStrm).Clear;

             tmpStrm.Position := 0;
             InStream.Position := 0;
             fOnEncryptStream(InStream, tmpstrm, fPassword);
             NewFile.CopyFrom(TmpStrm, 0);
             FileInfo.esize := tmpstrm.Size;
             FileInfo.Status := fsStoredEncrypted;
             FileInfo.CompressionLevel := fcNone;
           end
           else
           begin
             fOnEncryptStream(tmpstrm, Newfile, fpassword);
             FileInfo.CSize := TmpStrm.Size;
             FileInfo.esize := NewFile.Position - InternalFilePos(FileInfo.start);
             FileInfo.Status := fsCompressedEncrypted;
             FileInfo.CompressionLevel := fCompressionLevel;
           end;
        end
        else
        begin
           if TmpStrm.Size >= InStream.size then
           begin
             NewFile.CopyFrom(InStream, 0);
             FileInfo.Status := fsStored;
             FileInfo.CompressionLevel := fcNone;
           end
           else
           begin
             NewFile.CopyFrom(TmpStrm, 0);
             FileInfo.CSize := TmpStrm.Size;
             FileInfo.Status := fsCompressed;
             FileInfo.CompressionLevel := fCompressionLevel;
           end;
        end;

      finally
        TmpStrm.free;
        if not MemStrm then
        begin
          DeleteFile(TmpFile);
          tmpfile := '';
        end;
      end;
      UpdateRegistryRec(FileIndex, FileInfo);

      if assigned(fOnFileAddUpdate) then
        fOnFileAddUpdate(FileInfo.path, FileInfo.Name);

      for loop := FileIndex + 1 to FileCount - 1 do
      begin
        FileInfo := files[loop];
        FileInfo.start := NewFile.Position - fZeroPos;
        UpdateRegistryRec(loop, FileInfo);

        if FileInfo.eSize > 0 then
           wSize := FileInfo.eSize
        else if fileInfo.csize > 0 then
           wSize := FileInfo.cSize
        else
           wSize := fileinfo.osize;

        if wSize > 0 then
          NewFile.CopyFrom(ArchiveFile, wSize);
      end;

      WriteRegistry(NewFile);

    finally
      ArchiveFile.free;
    end;
    NewFile.Free;
    DeleteFile(fArchive);

    fProgressFlag := ifMovingTempFile;
    fProgressText := MovingTempFile;
    InternalProgress(self);

    RenameFile(TmpFileName, fArchive);
    TmpFileName := fArchive;
    CloseArchive;
    OpenArchive(TmpFileName);
  except
    DeleteFile(TmpFileName)
  end;
end;

procedure TZLBArc2.DeleteFiles(FileList: TStrings);
var
  TmpFile, OriginalFile: TStream;
  loop, loop2: longint;
  wSize : int64;
  filefound: boolean;
  TmpStr: string;
  TmpFileName: string;
  FileRec: TZLBFileRec2;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  if fVersionReadonly then
    raise EVersionError.create(OlderArchiveFormat);

  loop := 0;
  while loop < filecount do
  begin
    loop2 := 0;
    filefound := false;
    while loop2 < filelist.count do
    begin
      TmpStr := extractfilepath(filelist[loop2]);
      delete(TmpStr, 1, pos(':', TmpStr));
      TmpStr := TmpStr + extractfilename(filelist[loop2]);
      if lowercase(files[loop].path + files[loop].name) = lowercase(TmpStr) then
      begin
        filefound := true;
        break;
      end;
      inc(loop2);
      filefound := false;
    end;
    if filefound then
    begin
      filelist.delete(loop2);
      Freemem(fArchiveRegistry[loop], sizeof(TZLBFileRec2));
      fArchiveRegistry.Delete(loop);
    end
    else
      inc(loop);
  end;

  TmpFileName := GetTmpFileName;
  TmpFile := TFileStream.Create(TmpFileName, fmOpenReadWrite + fmShareExclusive);
  try
    try
      OriginalFile := TFileStream.Create(fArchive, fmOpenRead + fmShareDenyWrite);
      try
        if filecount > 0 then
        begin
          loop := 0;
          while loop < filecount do
          begin
            FileRec := files[loop];

            OriginalFile.Seek(InternalFilePos(FileRec.start), soFromBeginning);

            FileRec.Start := TmpFile.Position - fZeroPos;

            if FileRec.eSize > 0 then
               wSize := FileRec.eSize
            else if FileRec.csize > 0 then
               wSize := FileRec.cSize
            else
               wSize := FileRec.osize;

            if wSize > 0 then
               TmpFile.CopyFrom(OriginalFile, wSize);
               
            UpdateRegistryRec(loop, filerec);

            inc(loop);
          end;
        end;
      finally
        OriginalFile.Free;
      end;
    except
      raise EGeneralArchiveException.create(format(UnableToOpenFile, [fArchive]));
    end;

    WriteRegistry(TmpFile);

    TmpFile.Free;
    DeleteFile(fArchive);

    fProgressFlag := ifMovingTempFile;
    fProgressText := MovingTempFile;
    InternalProgress(self);

    RenameFile(TmpFileName, fArchive);
  finally
    TmpFileName := fArchive;
    CloseArchive;
    OpenArchive(TmpFileName);
    fProgressFlag := ifClear;
    InternalProgress(self);
  end;
end;

procedure TZLBArc2.CreateArchive(filename: string);
begin
  CloseArchive;
  fArchive := FileName;
  if (FileName = '') then
  begin
    fArchive := '';
    exit;
  end;
  if fileexists(filename) then
    raise EGeneralArchiveException.create(FileAlreadyExists);
  fNewArchive := true;
  if assigned(fOnArchiveOpened) then
    fOnArchiveOpened(self);
  fZeroPos := 0;
end;

function TZLBArc2.FileInArchive(filename: string): longint;
var
  loop: longint;
  tmpstr: string;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  result := -1;
  loop := 0;

  TmpStr := extractfilepath(filename);
  if SavePaths then
    delete(TmpStr, 1, pos(':', TmpStr));
  TmpStr := TmpStr + extractfilename(filename);

  while loop < fArchiveRegistry.count do
  begin
    if lowercase(Files[loop].Path + Files[loop].Name) = lowercase(tmpstr) then
    begin
      result := loop;
      break;
    end;
    inc(loop);
  end;
end;

function TZLBArc2.GetTmpFileName: string;
var
  tmpfile, TmpPath: string;

begin
  TmpPath := TempWorkingFolder;
  setlength(tmpfile, max_path);
  GetTempFilename(pchar(tmppath), 'ZLB', 0, pchar(tmpfile));
  result := tmpfile;
end;

procedure TZLBArc2.InternalProgress(Sender: TObject);
begin
    case fProgressFlag of
      ifOpening: if assigned(fOnProgress) then
          fOnprogress(fProgressText, round((fprogressIndex / fProgressTotal) * 100));
      ifExtracting: if assigned(fOnProgress) then
          fOnprogress(fProgressText, round((fprogressIndex / fProgressTotal) * 100));
      ifAdding: if assigned(fOnProgress) then
          fOnprogress(fProgressText, round((TCustomZStream(Sender).Position / fProgressTotal) * 100));
      ifUpdating: if assigned(fOnProgress) then
          fOnprogress(fProgressText, round((TCustomZStream(Sender).Position / fProgressTotal) * 100));
      ifWritingRegistry: if assigned(fOnProgress) then
          fOnprogress(fProgressText, round((fprogressIndex / fProgressTotal) * 100));
      ifClear: if assigned(fOnProgress) then
          fOnprogress(fArchive, 0);
      ifVerifyingArchiveCRC: if assigned(fOnProgress) then
          fOnProgress(fProgressText, round((fprogressIndex / fProgressTotal) * 100));
      ifMovingTempFile: if assigned(fOnProgress) then
          fOnProgress(fProgressText, 0);
    end;
end;

procedure TZLBArc2.UpdateRegistryRec(FileIndex: longint;
  FileInfo: TZLBFileRec2);
begin
  TZLBFileRec2(fArchiveRegistry[FileIndex]^) := FileInfo;
end;

procedure TZLBArc2.WriteRegistry(DataStrm: TStream);
var
  TmpStrm: TStream;
  loop: longint;
  ArchiveRec: TZLBArchiveRec2;
  ZStream: TCustomZStream;
begin
  TmpStrm := tmemorystream.Create;
  try
    try
      Fillchar(ArchiveRec, sizeof(TZLBArchiveRec2), #0);
      ArchiveRec.ArchiveSize := DataStrm.Size - fZeroPos;
      tmemorystream(TmpStrm).SetSize(sizeof(TZLBFileRec2) * fArchiveRegistry.count);
      for loop := 0 to fArchiveRegistry.count - 1 do
        TmpStrm.WriteBuffer(TZLBFileRec2(fArchiveRegistry.items[loop]^), sizeof(TZLBFileRec2));
      ZStream := TZCompressionStream.Create(DataStrm, TZCompressionLevel(fCompressionLevel));
      try
        fProgressFlag := ifWritingRegistry;
        fProgressTotal := TmpStrm.Size;
        fProgressIndex := 0;
        fProgressText := WritingRegistry;
        TZCompressionStream(ZStream).OnProgress := InternalProgress;
        ZStream.CopyFrom(TmpStrm, 0);
      finally
        ZStream.Free;
      end;
      ArchiveRec.RegistrySize := (DataStrm.size-fzeropos) - ArchiveRec.ArchiveSize;
      ArchiveRec.CompressedRegistry := true;
    except
      for loop := 0 to fArchiveRegistry.count - 1 do
        DataStrm.Writebuffer(TZLBFileRec2(fArchiveRegistry.items[loop]^), sizeof(TZLBFileRec2));
      ArchiveRec.CompressedRegistry := false;
    end;
    ArchiveRec.FileCount := fArchiveRegistry.count;
    ArchiveRec.Version := 2;
    DataStrm.position := 0;
    ArchiveRec.CRC := GetCRC32(DataStrm, 0, InternalFilePos(0));
    DataStrm.Seek(0, soFromEnd);
    DataStrm.WriteBuffer(ArchiveRec, SizeOf(TZLBArchiveRec2));
  finally
    TmpStrm.free;
  end;
end;

procedure TZLBArc2.RenameFileByName(Original, New: string);
var
  ArchiveFile, NewFile: TStream;
  TmpFileName: string;
  FileInfo: TZLBFileRec2;
  index: integer;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  if fVersionReadonly then
    raise EVersionError.create(OlderArchiveFormat);

  index := fileinarchive(original);
  if index <> -1 then
  begin
    TmpFileName := GetTmpFileName;
    NewFile := TFileStream.Create(TmpFileName, fmOpenReadWrite + fmShareExclusive);
    try
      if not fnewarchive then
      begin
        try
          ArchiveFile := TFileStream.Create(fArchive, fmOpenRead + fmShareDenyWrite);
        except
          raise EGeneralArchiveException.create(format(UnableToOpenFile, [fArchive]));
        end;
        try
          ArchiveFile.Position := 0;
          NewFile.CopyFrom(ArchiveFile, fArchiveDetails.ArchiveSize);
        finally
          ArchiveFile.Free;
        end;
      end;

      FileInfo := Files[index];
      FileInfo.name := extractfilename(new);
      FileInfo.path := extractfilepath(new);
      if FileInfo.path[length(FileInfo.path)] <> '\' then
        FileInfo.path := FileInfo.path + '\';
      UpdateRegistryRec(index, FileInfo);

      WriteRegistry(NewFile);

      NewFile.Free;
      DeleteFile(fArchive);

      fProgressFlag := ifMovingTempFile;
      fProgressText := MovingTempFile;
      InternalProgress(self);

      RenameFile(TmpFileName, fArchive);
      TmpFileName := fArchive;
      CloseArchive;
      OpenArchive(TmpFileName);
    except
      DeleteFile(TmpFileName)
    end;
  end;
end;

procedure TZLBArc2.GetFileList(FileList: TStrings);
var
  loop: integer;
begin
  filelist.clear;
  for loop := 0 to GetFileCount - 1 do
  begin
    with GetFileItem(loop) do
      filelist.Add(path + name);
  end;
end;

function TZLBArc2.GetTempFolder: string;
var
  tmppath: string;
begin
  if FTempFolder = '' then
  begin
    setlength(tmppath, max_path);
    GetTempPath(max_path, pchar(tmppath));
    result := TmpPath;
  end
  else
    result := fTempFolder;
end;

function TZLBArc2.InternalFilePos(ReportedPos: Int64): Int64;
begin
   if fZeroPos > -1 then
      result := fZeroPos + ReportedPos
   else
      raise EFileCorruption.create('Unknown file position marker');
end;

procedure TZLBArc2.ConvertArchive(NewFilename:string);
var
  ArchiveFile, NewFile : TStream;
begin
  if fArchive = '' then
    raise ENoArchive.create(NoArchiveOpen);

  if not fVersionReadonly then
    raise EVersionError.create(AlreadyCurrentArchiveFormat);

  if FileExists(NewFileName) then
     raise EZLBArchiveException.create(ArchiveFileAlreadyExists);

  NewFile := TFileStream.Create(NewFilename, fmCreate);
  try
     try
       ArchiveFile := TFileStream.Create(fArchive, fmOpenRead + fmShareDenyWrite);
     except
       raise EGeneralArchiveException.create(format(UnableToOpenFile, [fArchive]));
     end;

     try
       ArchiveFile.Position := 0;
       NewFile.CopyFrom(ArchiveFile, InternalFilePos(fArchiveDetails.ArchiveSize));
     finally
       ArchiveFile.Free;
     end;

     WriteRegistry(NewFile);
  finally
     NewFile.free;
  end;
end;

end.

